@@ -2,6 +2,6 @@
 #define OPR_H 1
 
 #define opr_containerof(ptr, structure, member) \
-	((structure *)((char *)(ptr)-(char *)(&((structure *)NULL)->member)))
+	((structure *)((char *)(ptr)-(char *)offsetof(structure, member)))
 
 #endif				/* OPR_H */
