@@ -51,6 +51,10 @@
 #include <rpc/svc_auth.h>
 #include "svc_ioq.h"
 
+#ifdef USE_RPC_RDMA
+#include "rpc_rdma.h"
+#endif // USE_RPC_RDMA
+
 /**
  * @file svc_rqst.c
  * @contributeur William Allen Simpson <bill@cohortfs.com>
@@ -521,9 +525,9 @@ svc_rqst_unhook_events(struct rpc_dplx_rec *rec, struct svc_rqst_rec *sr_rec,
 {
 	int code = EINVAL;
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, unhook, __func__, __LINE__, &rec->xprt, ev_flags);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_AUTO_TRACEPOINT(&rec->xprt, unhook, TRACE_DEBUG,
+		"Unhook. ev_flags: {}", ev_flags);
+
 	__warnx(TIRPC_DEBUG_FLAG_SVC_RQST,
 		"%s: xprt %p fd %d ev_flags%s%s%s%s%s%s%s%s%s",
 		__func__, &rec->xprt, rec->xprt.xp_fd,
@@ -639,9 +643,13 @@ svc_rqst_rearm_events_locked(SVCXPRT *xprt, uint16_t ev_flags)
 	struct svc_rqst_rec *sr_rec = rec->ev_p;
 	int code = EINVAL;
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, rearm, __func__, __LINE__, xprt, ev_flags);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_AUTO_TRACEPOINT(xprt, rearm, TRACE_DEBUG,
+		"Rearm. ev_flags: {}", ev_flags);
+
+	const bool is_xprt_destroyed = xprt->xp_flags & (ev_flags | SVC_XPRT_FLAG_DESTROYED);
+	/* MUST follow the destroyed check above */
+	const bool is_rec_shutdown = !is_xprt_destroyed && (
+		sr_rec->ev_flags & SVC_RQST_FLAG_SHUTDOWN);
 
 	__warnx(TIRPC_DEBUG_FLAG_SVC_RQST,
 		"%s: xprt %p fd %d ev_flags%s%s%s%s%s%s%s%s%s",
@@ -654,14 +662,9 @@ svc_rqst_rearm_events_locked(SVCXPRT *xprt, uint16_t ev_flags)
 		ev_flags & SVC_XPRT_FLAG_DESTROYING  ? " DESTROYING" : "",
 		ev_flags & SVC_XPRT_FLAG_RELEASING   ? " RELEASING" : "",
 		ev_flags & SVC_XPRT_FLAG_UREG        ? " UREG" : "",
-		sr_rec->ev_flags & SVC_RQST_FLAG_SHUTDOWN
-					? "sr_rec->ev_flags SHUTDOWN" : "");
+		is_rec_shutdown                      ? "sr_rec->ev_flags SHUTDOWN" : "");
 
-	if (xprt->xp_flags & (ev_flags | SVC_XPRT_FLAG_DESTROYED))
-		return (0);
-
-	/* MUST follow the destroyed check above */
-	if (sr_rec->ev_flags & SVC_RQST_FLAG_SHUTDOWN)
+	if (is_xprt_destroyed || is_rec_shutdown)
 		return (0);
 
 	/* Don't take a ref on the xprt.  We take a ref in hook, and release it
@@ -774,9 +777,8 @@ svc_rqst_hook_events(struct rpc_dplx_rec *rec, struct svc_rqst_rec *sr_rec,
 {
 	int code = EINVAL;
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, hook, __func__, __LINE__, &rec->xprt, ev_flags);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_AUTO_TRACEPOINT(&rec->xprt, hook, TRACE_DEBUG,
+		"Hook. ev_flags: {}", ev_flags);
 
 	__warnx(TIRPC_DEBUG_FLAG_SVC_RQST,
 		"%s: xprt %p fd %d ev_flags%s%s%s%s%s%s%s%s%s",
@@ -1156,6 +1158,37 @@ svc_rqst_xprt_unregister(SVCXPRT *xprt, uint32_t flags)
 	svc_rqst_unreg(rec, sr_rec);
 }
 
+#if defined(USE_RPC_RDMA)
+
+/*
+ * flags indicate locking state
+ *
+ * @note Locking
+ *	Called via svc_release_it() with once-only semantic.
+ */
+void
+svc_rqst_xprt_unregister_rdma(SVCXPRT *xprt, uint32_t flags)
+{
+	struct rpc_dplx_rec *rec = REC_XPRT(xprt);
+	struct svc_rqst_rec *sr_rec = (struct svc_rqst_rec *)rec->ev_p;
+
+	assert(!sr_rec);
+
+	/* Remove from the transport list here (and only here)
+	 * before clearing the registration to ensure other
+	 * lookups cannot re-use this transport.
+	 */
+	if (!(flags & RPC_DPLX_LOCKED))
+		rpc_dplx_rli(rec);
+
+	svc_xprt_clear(xprt);
+
+	if (!(flags & RPC_DPLX_LOCKED))
+		rpc_dplx_rui(rec);
+}
+
+#endif
+
 /*static*/ void
 svc_rqst_xprt_task_recv(struct work_pool_entry *wpe)
 {
@@ -1165,12 +1198,7 @@ svc_rqst_xprt_task_recv(struct work_pool_entry *wpe)
 
 	atomic_clear_uint16_t_bits(&ioq->ioq_s.qflags, IOQ_FLAG_WORKING);
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, recv, __func__, __LINE__,
-		   &rec->xprt,
-		   (unsigned int)(rec->xprt.xp_flags & SVC_XPRT_FLAG_DESTROYED),
-		   (unsigned int) rec->xprt.xp_refcnt);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_AUTO_TRACEPOINT(&rec->xprt, task_recv, TRACE_DEBUG, "Hook");
 
 	/* atomic barrier (above) should protect following values */
 	if (rec->xprt.xp_refcnt > 1
@@ -1258,12 +1286,8 @@ svc_rqst_xprt_task_send(struct work_pool_entry *wpe)
 
 	atomic_clear_uint16_t_bits(&ioq->ioq_s.qflags, IOQ_FLAG_WORKING);
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, send, __func__, __LINE__,
-		   &rec->xprt,
-		   (unsigned int)(rec->xprt.xp_flags & SVC_XPRT_FLAG_DESTROYED),
-		   (unsigned int) rec->xprt.xp_refcnt);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_AUTO_TRACEPOINT(&rec->xprt, task_send, TRACE_DEBUG,
+		"Task send");
 
 	/* atomic barrier (above) should protect following values */
 	if (rec->xprt.xp_refcnt > 1
@@ -1419,10 +1443,8 @@ svc_rqst_epoll_event(struct svc_rqst_rec *sr_rec, struct epoll_event *ev)
 		ev_flag & SVC_XPRT_FLAG_ADDED_RECV ? " ADDED_RECV" : "",
 		ev_flag & SVC_XPRT_FLAG_ADDED_SEND ? " ADDED_SEND" : "");
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, event, __func__, __LINE__, &rec->xprt, xp_flags,
-		   ev_flag);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_AUTO_TRACEPOINT(&rec->xprt, epoll_event, TRACE_DEBUG,
+		"Epoll event. ev_flag: {}", ev_flag);
 
 	if (rec->xprt.xp_refcnt > 1
 	    && (xp_flags & ev_flag)
@@ -1682,3 +1704,16 @@ svc_rqst_shutdown(void)
 		svc_rqst_delete_evchan(--channels);
 	}
 }
+
+int
+svc_get_port(sockaddr_t *addr)
+{
+	switch (addr->ss_family) {
+	case AF_INET:
+		return ntohs(((struct sockaddr_in *)addr)->sin_port);
+	case AF_INET6:
+		return ntohs(((struct sockaddr_in6 *)addr)->sin6_port);
+	default:
+		return -1;
+	}
+}
