@@ -428,9 +428,8 @@ svc_vc_rendezvous(SVCXPRT *xprt)
 	static int n = 1;
 	struct timeval timeval;
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, funcin, __func__, __LINE__, xprt);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_AUTO_TRACEPOINT(xprt, rendezvous_start, TRACE_INFO,
+		"rendezvous_start");
 
  again:
 	len = sizeof(addr);
@@ -533,13 +532,19 @@ svc_vc_rendezvous(SVCXPRT *xprt)
 	newxprt->xp_parent = xprt;
 	if (xprt->xp_dispatch.rendezvous_cb(newxprt)
 	 || svc_rqst_xprt_register(newxprt, xprt)) {
+		// Note xp_parent is released in svc_vc_destroy_task
 		SVC_DESTROY(newxprt);
 		/* Was never added to epoll */
 		SVC_RELEASE(newxprt, SVC_RELEASE_FLAG_NONE);
-		SVC_RELEASE(xprt, SVC_RELEASE_FLAG_NONE);
 		return (XPRT_DESTROYED);
 	}
 
+	__warnx(TIRPC_DEBUG_FLAG_SVC_VC,
+		"New client connected "
+		"xprt %p, fd %d, port %d",
+		newxprt, newxprt->xp_fd,
+		svc_get_port(newxprt->xp_local.nb.buf));
+
 	/* We're not using a ref for the hook anymore, since epoll doesn't store
 	 * the transport pointer.  Drop the extra ref here. */
 	SVC_RELEASE(newxprt, SVC_RELEASE_FLAG_NONE);
@@ -552,39 +557,55 @@ svc_vc_destroy_task(struct work_pool_entry *wpe)
 	struct rpc_dplx_rec *rec =
 			opr_containerof(wpe, struct rpc_dplx_rec, ioq.ioq_wpe);
 	uint16_t xp_flags;
-	bool reset_xprt_fd = false;
+	bool close_fd = false;
 
+	const int32_t xp_refcnt = atomic_fetch_int32_t(&rec->xprt.xp_refcnt);
 	__warnx(TIRPC_DEBUG_FLAG_REFCNT,
 		"%s() %p fd %d xp_refcnt %" PRId32,
-		__func__, rec, rec->xprt.xp_fd, rec->xprt.xp_refcnt);
+		__func__, rec, rec->xprt.xp_fd, xp_refcnt);
 
-	if (rec->xprt.xp_refcnt) {
+	if (xp_refcnt > 0) {
 		/* instead of nanosleep */
 		work_pool_submit(&svc_work_pool, &(rec->ioq.ioq_wpe));
 		return;
+	} else if (unlikely(xp_refcnt < 0)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s() negative refcnt: %p fd %d xp_refcnt %" PRId32,
+			__func__, rec, rec->xprt.xp_fd, xp_refcnt);
+		abort();
 	}
 
 	xp_flags = atomic_postclear_uint16_t_bits(&rec->xprt.xp_flags,
 						  SVC_XPRT_FLAG_CLOSE);
-	if ((xp_flags & SVC_XPRT_FLAG_CLOSE) && rec->xprt.xp_fd != RPC_ANYFD) {
-		(void)close(rec->xprt.xp_fd);
+	close_fd = ((xp_flags & SVC_XPRT_FLAG_CLOSE) &&
+		rec->xprt.xp_fd != RPC_ANYFD);
+	if (close_fd) {
+		/* Shutting down without releasing the fd, since
+		 * xp_free_user_data() might be using it */
+		(void)shutdown(rec->xprt.xp_fd, SHUT_RDWR);
 		__warnx(TIRPC_DEBUG_FLAG_SVC_VC,
-			"%s: fd %d closed",
+			"%s: fd %d shutdown",
 			 __func__, rec->xprt.xp_fd);
-		/* Mark xprt_fd for reset */
-		reset_xprt_fd = true;
-		rec->xprt.xp_fd = RPC_ANYFD;
 		if (rec->xprt.xp_fd_send != RPC_ANYFD)
-			(void)close(rec->xprt.xp_fd_send);
-		rec->xprt.xp_fd_send = RPC_ANYFD;
+			(void)shutdown(rec->xprt.xp_fd_send, SHUT_RDWR);
 	}
 
 	if (rec->xprt.xp_ops->xp_free_user_data)
 		rec->xprt.xp_ops->xp_free_user_data(&rec->xprt);
 
-	/* Reset xprt's FD after the xp_free_user_data call */
-	if (reset_xprt_fd) {
+	/* Close and reset xprt's FD after the xp_free_user_data call.
+	 * It's safe to release the FD at this point (by calling close), since
+	 * there are no references left to this XPRT. */
+	if (close_fd) {
+		__warnx(TIRPC_DEBUG_FLAG_SVC_VC,
+			"%s: fd %d close",
+			 __func__, rec->xprt.xp_fd);
+		(void)close(rec->xprt.xp_fd);
 		rec->xprt.xp_fd = RPC_ANYFD;
+		if (rec->xprt.xp_fd_send != RPC_ANYFD) {
+			(void)close(rec->xprt.xp_fd_send);
+			rec->xprt.xp_fd_send = RPC_ANYFD;
+		}
 	}
 
 	if (rec->xprt.xp_tp)
@@ -712,6 +733,209 @@ svc_vc_stat(SVCXPRT *xprt)
 	return (XPRT_IDLE);
 }
 
+static bool is_remote_addr_set(SVCXPRT *xprt)
+{
+	return (xprt->xp_flags & SVC_XPRT_FLAG_REMOTE_ADDR_SET);
+}
+
+static bool update_and_notify_remote_address_set(SVCXPRT *xprt)
+{
+	u_int prev_xp_flags;
+	__warnx(TIRPC_DEBUG_FLAG_SVC_VC,
+			"%s: %p fd %d update remote address set",
+			__func__, xprt, xprt->xp_fd);
+	prev_xp_flags = atomic_postset_uint16_t_bits(&xprt->xp_flags,
+			SVC_XPRT_FLAG_REMOTE_ADDR_SET);
+	/* remote addr set was must be called only once for xprt */
+	assert(!(prev_xp_flags & SVC_XPRT_FLAG_REMOTE_ADDR_SET));
+	if (xprt->xp_dispatch.remote_addr_set_cb) {
+		if (xprt->xp_dispatch.remote_addr_set_cb(xprt)) {
+			__warnx(TIRPC_DEBUG_FLAG_ERROR,
+				"%s: %p fd %d remote_addr_set cb failed"
+				" (will set dead)",
+				__func__, xprt, xprt->xp_fd);
+			return false;
+		}
+	}
+	return true;
+}
+
+enum haproxy_ret_code {
+       HAPROXY_RET_CODE__SUCCESS = 0,
+       HAPROXY_RET_CODE__FAILURE,
+       HAPROXY_RET_CODE__IGNORE_LOCAL,
+       HAPROXY_RET_CODE__NOT_HAPROXY
+};
+
+static enum haproxy_ret_code handle_haproxy_header(SVCXPRT *xprt)
+{
+	/* HA Proxy V2? */
+	ssize_t rlen;
+	uint32_t rest[2];
+	struct proxy_header_part s;
+	union proxy_addr pa;
+	enum haproxy_ret_code ret;
+
+	__warnx(TIRPC_DEBUG_FLAG_SVC_VC,
+			"%s: %p fd %d potential haproxy packet",
+			__func__, xprt, xprt->xp_fd);
+
+	/* PEEK in order not to consume a non haproxy packet */
+	rlen = recv(xprt->xp_fd, rest, sizeof(rest), MSG_WAITALL | MSG_PEEK);
+	if (rlen != sizeof(rest)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: %p fd %d proxy header failed rest rlen = %z "
+			"(will set dead)",
+			__func__, xprt, xprt->xp_fd, rlen);
+		return HAPROXY_RET_CODE__FAILURE;
+	}
+	rest[0] = ntohl(rest[0]);
+	rest[1] = ntohl(rest[1]);
+
+	if (rest[0] != PP2_SIG_UINT32_2 || rest[1] != PP2_SIG_UINT32_3) {
+		__warnx(TIRPC_DEBUG_FLAG_WARN,
+			"%s: %p fd %d proxy header failed rest1=%08x rest2=%08x "
+			"(treat as regular rpc packet)",
+			__func__, xprt, xprt->xp_fd, (int) rest[0],
+			(int) rest[1]);
+		/* The signature does not fully match.
+		 * Flow should treat the packet as a regular rpc packet.*/
+		return HAPROXY_RET_CODE__NOT_HAPROXY;
+	}
+
+	rlen = recv(xprt->xp_fd, rest, sizeof(rest), MSG_WAITALL);
+	if (rlen != sizeof(rest)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: %p fd %d proxy header failed rest rlen = %z "
+			"(will set dead)",
+			__func__, xprt, xprt->xp_fd, rlen);
+		return HAPROXY_RET_CODE__FAILURE;
+	}
+
+	rlen = recv(xprt->xp_fd, &s, sizeof(s), MSG_WAITALL);
+
+	if (rlen != sizeof(s)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: %p fd %d proxy header failed header rlen = %z "
+			"(will set dead)",
+			__func__, xprt, xprt->xp_fd, rlen);
+		return HAPROXY_RET_CODE__FAILURE;
+	}
+
+	s.len = ntohs(s.len);
+	if (unlikely(s.len > sizeof(pa))) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: %p fd %d incorrect proxy header "
+			"addr len = %z (will set dead)",
+			__func__, xprt, xprt->xp_fd, s.len);
+		return HAPROXY_RET_CODE__FAILURE;
+	}
+
+	rlen = recv(xprt->xp_fd, &pa, s.len, MSG_WAITALL);
+
+	if (rlen != s.len) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: %p fd %d proxy header rest len failed header "
+			"rlen = %z (will set dead)",
+			__func__, xprt, xprt->xp_fd, rlen);
+		return HAPROXY_RET_CODE__FAILURE;
+	}
+
+	if (s.ver_cmd == PP2_VERSIOB2_CMD_PROXY) {
+		if (unlikely(is_remote_addr_set(xprt))) {
+			/* We don't allow more than one proxy protocol packet.
+			   Allowing it will cause a security vulnerability where
+			   at any point the client could sent a PP packet and
+			   change its IP to circumvent any IP based access rules.
+			*/
+			__warnx(TIRPC_DEBUG_FLAG_WARN,
+				"%s: %p fd %d got more than one PP packet. "
+				"This is not allowed - terminating",
+				__func__, xprt, xprt->xp_fd);
+			return HAPROXY_RET_CODE__FAILURE;
+		}
+		if (s.fam == PP2_TRANS_STREAM_FAM_INET) {
+			if (unlikely(s.len < sizeof(pa.ip4))) {
+				__warnx(TIRPC_DEBUG_FLAG_ERROR,
+					"%s: %p fd %d incorrect proxy header "
+					"ipv4 addr len = %z (will set dead)",
+					__func__, xprt, xprt->xp_fd, s.len);
+				return HAPROXY_RET_CODE__FAILURE;
+			}
+			struct sockaddr_in *ss4;
+
+			xprt->xp_proxy = xprt->xp_remote;
+			ss4 = (struct sockaddr_in *)
+					&xprt->xp_remote.ss;
+			ss4->sin_family = AF_INET;
+			memcpy(&ss4->sin_addr,
+			       &pa.ip4.src_addr,
+			       sizeof(struct in_addr));
+			ss4->sin_port = pa.ip4.src_port;
+
+		} else if (s.fam ==
+				   PP2_TRANS_STREAM_FAM_INET6) {
+			if (unlikely(s.len < sizeof(pa.ip6))) {
+				__warnx(TIRPC_DEBUG_FLAG_ERROR,
+					"%s: %p fd %d incorrect proxy header "
+					"ipv6 addr len = %z (will set dead)",
+					__func__, xprt, xprt->xp_fd, s.len);
+				return HAPROXY_RET_CODE__FAILURE;
+			}
+			struct sockaddr_in6 *ss6;
+
+			xprt->xp_proxy = xprt->xp_remote;
+			ss6 = (struct sockaddr_in6 *)
+					&xprt->xp_remote.ss;
+			xprt->xp_remote.ss.ss_family = AF_INET6;
+			memcpy(&ss6->sin6_addr,
+			       &pa.ip6.src_addr,
+			       sizeof(struct in6_addr));
+			ss6->sin6_port = pa.ip6.src_port;
+
+		} else {
+			/* NOTE: we don't support UNIX or UDP
+			 * sockets
+			 */
+			__warnx(TIRPC_DEBUG_FLAG_ERROR,
+				"%s: %p fd %d invalid proxy protocol = %0x2 "
+				"(will set dead)",
+				__func__, xprt, xprt->xp_fd, (int) s.fam);
+			return HAPROXY_RET_CODE__FAILURE;
+		}
+
+		return HAPROXY_RET_CODE__SUCCESS;
+
+	} else if (s.ver_cmd == PP2_VERSION2_CMD_LOCAL) {
+		__warnx(TIRPC_DEBUG_FLAG_EVENT,
+			"%s: %p fd %d proxy ignored for local",
+			__func__, xprt, xprt->xp_fd);
+		ret = HAPROXY_RET_CODE__IGNORE_LOCAL;
+	} else {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: %p fd %d invalid proxy command = %0x2 (will set dead)",
+			__func__, xprt, xprt->xp_fd,(int) s.ver_cmd);
+		return HAPROXY_RET_CODE__FAILURE;
+	}
+
+	if (unlikely(svc_rqst_rearm_events(xprt,
+				   SVC_XPRT_FLAG_ADDED_RECV))) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: %p fd %d svc_rqst_rearm_events failed (will set dead)",
+			__func__, xprt, xprt->xp_fd);
+		ret = HAPROXY_RET_CODE__FAILURE;
+
+		XPRT_AUTO_TRACEPOINT(xprt, rearm_failed,
+			TRACE_ERR, "Rearm failed");
+	} else {
+		XPRT_UNIQUE_AUTO_TRACEPOINT(
+			xprt, recv_exit, TRACE_DEBUG,
+			"Exiting recv");
+	}
+
+	return ret;
+}
+
 static enum xprt_stat
 svc_vc_recv(SVCXPRT *xprt)
 {
@@ -725,9 +949,7 @@ svc_vc_recv(SVCXPRT *xprt)
 	int code;
 	bool hap_again = false;
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, funcin, __func__, __LINE__, xprt);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_AUTO_TRACEPOINT(xprt, recv_start, TRACE_DEBUG, "recv_start");
 
 	/* no need for locking, only one svc_rqst_xprt_task() per event.
 	 * depends upon svc_rqst_rearm_events() for ordering.
@@ -766,20 +988,17 @@ again:
 					SVC_DESTROY(xprt);
 					code = EINVAL;
 				}
-#ifdef USE_LTTNG_NTIRPC
-				tracepoint(xprt, recv_exit, __func__, __LINE__,
-					   xprt, "EAGAIN", code);
-#endif /* USE_LTTNG_NTIRPC */
+				XPRT_AUTO_TRACEPOINT(xprt, recv_eagin,
+					TRACE_DEBUG, "recv got EAGAIN");
 				return SVC_STAT(xprt);
 			}
 			__warnx(TIRPC_DEBUG_FLAG_WARN,
 				"%s: %p fd %d recv errno %d (will set dead)",
 				"svc_vc_wait", xprt, xprt->xp_fd, code);
 			SVC_DESTROY(xprt);
-#ifdef USE_LTTNG_NTIRPC
-			tracepoint(xprt, recv_exit, __func__, __LINE__,
-				   xprt, "ERROR", code);
-#endif /* USE_LTTNG_NTIRPC */
+
+			XPRT_AUTO_TRACEPOINT(xprt, recv_err,
+				TRACE_WARNING, "recv got errno: {}", code);
 			return SVC_STAT(xprt);
 		}
 
@@ -788,10 +1007,10 @@ again:
 				"%s: %p fd %d recv closed (will set dead)",
 				"svc_vc_wait", xprt, xprt->xp_fd);
 			SVC_DESTROY(xprt);
-#ifdef USE_LTTNG_NTIRPC
-			tracepoint(xprt, recv_exit, __func__, __LINE__,
-				   xprt, "EMPTY", 0);
-#endif /* USE_LTTNG_NTIRPC */
+
+			XPRT_AUTO_TRACEPOINT(xprt, recv_empty,
+				TRACE_DEBUG, "recv EOF");
+
 			return SVC_STAT(xprt);
 		}
 
@@ -802,128 +1021,26 @@ again:
 
 		if (xd->sx_fbtbc == PP2_SIG_UINT32) {
 			/* HA Proxy V2? */
-			uint32_t rest[2];
-			struct proxy_header_part s;
-			union proxy_addr *pa;
-
-			rlen = recv(xprt->xp_fd, rest, sizeof(rest),
-				    MSG_WAITALL);
-			if (rlen != sizeof(rest)) {
-				__warnx(TIRPC_DEBUG_FLAG_ERROR,
-					"%s: %p fd %d proxy header failed rest rlen = %z (will set dead)",
-				__func__, xprt, xprt->xp_fd, rlen);
-				SVC_DESTROY(xprt);
-				return SVC_STAT(xprt);
-			}
-			rest[0] = ntohl(rest[0]);
-			rest[1] = ntohl(rest[1]);
-
-			if (rest[0] != PP2_SIG_UINT32_2 ||
-			    rest[1] != PP2_SIG_UINT32_3) {
-				__warnx(TIRPC_DEBUG_FLAG_ERROR,
-					"%s: %p fd %d proxy header failed rest1=%08x rest2=%08x (will set dead)",
-				__func__, xprt, xprt->xp_fd, (int) rest[0], (int) rest[1]);
-				SVC_DESTROY(xprt);
-			}
-
-			rlen = recv(xprt->xp_fd, &s, sizeof(s),
-				    MSG_WAITALL);
-
-			if (rlen != sizeof(s)) {
-				__warnx(TIRPC_DEBUG_FLAG_ERROR,
-					"%s: %p fd %d proxy header failed header rlen = %z (will set dead)",
-				__func__, xprt, xprt->xp_fd, rlen);
-				SVC_DESTROY(xprt);
-				return SVC_STAT(xprt);
-			}
-
-			s.len = ntohs(s.len);
-			pa = mem_zalloc(s.len);
-			rlen = recv(xprt->xp_fd, pa, s.len, MSG_WAITALL);
-
-			if (rlen != s.len) {
-				__warnx(TIRPC_DEBUG_FLAG_ERROR,
-					"%s: %p fd %d proxy header rest len failed header rlen = %z (will set dead)",
-				__func__, xprt, xprt->xp_fd, rlen);
-				SVC_DESTROY(xprt);
-				return SVC_STAT(xprt);
-			}
-
-			if (s.ver_cmd == PP2_VERSIOB2_CMD_PROXY) {
-				if (s.fam == PP2_TRANS_STREAM_FAM_INET) {
-					struct sockaddr_in *ss4;
-
-					xprt->xp_proxy = xprt->xp_remote;
-					ss4 = (struct sockaddr_in *)
-							&xprt->xp_remote.ss;
-					ss4->sin_family = AF_INET;
-					memcpy(&ss4->sin_addr,
-					       &pa->ip4.src_addr,
-					       sizeof(struct in_addr));
-					ss4->sin_port = pa->ip4.src_port;
-
-				} else if (s.fam ==
-						   PP2_TRANS_STREAM_FAM_INET6) {
-					struct sockaddr_in6 *ss6;
-
-					xprt->xp_proxy = xprt->xp_remote;
-					ss6 = (struct sockaddr_in6 *)
-							&xprt->xp_remote.ss;
-					xprt->xp_remote.ss.ss_family = AF_INET6;
-					memcpy(&ss6->sin6_addr,
-					       &pa->ip6.src_addr,
-					       sizeof(struct in6_addr));
-					ss6->sin6_port = pa->ip6.src_port;
-
-				} else {
-					/* NOTE: we don't support UNIX or UDP
-					 * sockets
-					 */
-					__warnx(TIRPC_DEBUG_FLAG_ERROR,
-						"%s: %p fd %d invalid proxy protocol = %0x2 (will set dead)",
-						__func__, xprt, xprt->xp_fd,
-						(int) s.fam);
+			enum haproxy_ret_code ret = handle_haproxy_header(xprt);
+			switch (ret) {
+			case HAPROXY_RET_CODE__SUCCESS:
+				if (!update_and_notify_remote_address_set(xprt)) {
 					SVC_DESTROY(xprt);
 					return SVC_STAT(xprt);
 				}
-
 				/* Now look to see if there's more... */
 				hap_again = true;
 				goto again;
-
-			} else if (s.ver_cmd == PP2_VERSION2_CMD_LOCAL) {
-				__warnx(TIRPC_DEBUG_FLAG_EVENT,
-					"%s: %p fd %d proxy ignored for local",
-					__func__, xprt, xprt->xp_fd);
-			} else {
-				__warnx(TIRPC_DEBUG_FLAG_ERROR,
-					"%s: %p fd %d invalid proxy command = %0x2 (will set dead)",
-					__func__, xprt, xprt->xp_fd,
-					(int) s.ver_cmd);
+			case HAPROXY_RET_CODE__FAILURE:
 				SVC_DESTROY(xprt);
 				return SVC_STAT(xprt);
+			case HAPROXY_RET_CODE__IGNORE_LOCAL:
+				return SVC_STAT(xprt);
+			case HAPROXY_RET_CODE__NOT_HAPROXY:
+				break;
 			}
-
-			if (unlikely(svc_rqst_rearm_events(xprt,
-						   SVC_XPRT_FLAG_ADDED_RECV))) {
-				__warnx(TIRPC_DEBUG_FLAG_ERROR,
-					"%s: %p fd %d svc_rqst_rearm_events failed (will set dead)",
-				__func__, xprt, xprt->xp_fd);
-				SVC_DESTROY(xprt);
-#ifndef USE_LTTNG_NTIRPC
-			}
-#else
-				tracepoint(xprt, recv_exit, __func__, __LINE__,
-					   xprt, "REARM FAILED", -1);
-			} else {
-				tracepoint(xprt, recv_exit, __func__, __LINE__,
-					   xprt, "MORE", 0);
-			}
-#endif /* USE_LTTNG_NTIRPC */
-			return SVC_STAT(xprt);
-			/* return (XPRT_IDLE); */
 		}
-	
+
 		flags = UIO_FLAG_FREE | UIO_FLAG_MORE;
 
 		if (xd->sx_fbtbc & LAST_FRAG) {
@@ -936,17 +1053,14 @@ again:
 				"%s: %p fd %d fragment is zero (will set dead)",
 				__func__, xprt, xprt->xp_fd);
 			SVC_DESTROY(xprt);
-#ifdef USE_LTTNG_NTIRPC
-			tracepoint(xprt, recv_exit, __func__, __LINE__,
-				   xprt, "NO RECORD", 0);
-#endif /* USE_LTTNG_NTIRPC */
+
+			XPRT_AUTO_TRACEPOINT(xprt, recv_no_record,
+				TRACE_ERR,
+				"Recv fragmet is zero (will set dead)");
+
 			return SVC_STAT(xprt);
 		}
 
-#ifdef USE_LTTNG_NTIRPC
-		tracepoint(xprt, recv_frag, __func__, __LINE__,
-			   xprt, xd->sx_fbtbc);
-#endif /* USE_LTTNG_NTIRPC */
 		/* one buffer per fragment */
 		uv = xdr_ioq_uv_create(xd->sx_fbtbc, flags);
 		(xioq->ioq_uv.uvqh.qcount)++;
@@ -974,20 +1088,20 @@ again:
 				SVC_DESTROY(xprt);
 				code = EINVAL;
 			}
-#ifdef USE_LTTNG_NTIRPC
-			tracepoint(xprt, recv_exit, __func__, __LINE__,
-				   xprt, "EAGAIN", code);
-#endif /* USE_LTTNG_NTIRPC */
+
+			XPRT_AUTO_TRACEPOINT(xprt, recv_eagin2,
+					TRACE_DEBUG, "recv got EAGIN");
+
 			return SVC_STAT(xprt);
 		}
 		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s: %p fd %d recv errno %d (will set dead)",
 			__func__, xprt, xprt->xp_fd, code);
 		SVC_DESTROY(xprt);
-#ifdef USE_LTTNG_NTIRPC
-		tracepoint(xprt, recv_exit, __func__, __LINE__,
-			   xprt, "ERROR", code);
-#endif /* USE_LTTNG_NTIRPC */
+
+		XPRT_AUTO_TRACEPOINT(xprt, recv_error,
+				TRACE_ERR, "recv got errno: {}", code);
+
 		return SVC_STAT(xprt);
 	}
 
@@ -996,17 +1110,16 @@ again:
 			"%s: %p fd %d recv closed (will set dead)",
 			__func__, xprt, xprt->xp_fd);
 		SVC_DESTROY(xprt);
-#ifdef USE_LTTNG_NTIRPC
-		tracepoint(xprt, recv_exit, __func__, __LINE__,
-			   xprt, "EMPTY", 0);
-#endif /* USE_LTTNG_NTIRPC */
+
+		XPRT_AUTO_TRACEPOINT(xprt, recv_closed,
+				TRACE_INFO, "recv closed");
+
 		return SVC_STAT(xprt);
 	}
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, recv_bytes, __func__, __LINE__,
-		   xprt, xd->sx_fbtbc, rlen);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_AUTO_TRACEPOINT(xprt, recv_bytes,
+		TRACE_DEBUG, "recv {} bytes in sx_fbtbc: {}", rlen,
+			xd->sx_fbtbc);
 
 	uv->v.vio_tail += rlen;
 	xd->sx_fbtbc -= rlen;
@@ -1021,17 +1134,14 @@ again:
 			__warnx(TIRPC_DEBUG_FLAG_ERROR,
 				"%s: %p fd %d svc_rqst_rearm_events failed (will set dead)",
 				__func__, xprt, xprt->xp_fd);
+			XPRT_UNIQUE_AUTO_TRACEPOINT(xprt, rearm_failed,
+				TRACE_ERR, "Rearm failed");
 			SVC_DESTROY(xprt);
-#ifndef USE_LTTNG_NTIRPC
-		}
-#else
-			tracepoint(xprt, recv_exit, __func__, __LINE__,
-				   xprt, "REARM FAILED", -1);
 		} else {
-			tracepoint(xprt, recv_exit, __func__, __LINE__,
-				   xprt, "MORE", 0);
+			XPRT_UNIQUE_AUTO_TRACEPOINT(xprt, recv_exit,
+				TRACE_DEBUG, "recv exit");
 		}
-#endif /* USE_LTTNG_NTIRPC */
+
 		return SVC_STAT(xprt);
 	}
 
@@ -1040,23 +1150,29 @@ again:
 	TAILQ_REMOVE(&rec->ioq.ioq_uv.uvqh.qh, &xioq->ioq_s, q);
 	xdr_ioq_reset(xioq, 0);
 
+	if (!is_remote_addr_set(xprt)) {
+		if (!update_and_notify_remote_address_set(xprt)) {
+			SVC_DESTROY(xprt);
+			return SVC_STAT(xprt);
+		}
+	}
+
 	if (unlikely(svc_rqst_rearm_events(xprt, SVC_XPRT_FLAG_ADDED_RECV))) {
 		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s: %p fd %d svc_rqst_rearm_events failed (will set dead)",
 			__func__, xprt, xprt->xp_fd);
 		xdr_ioq_destroy(xioq, xioq->ioq_s.qsize);
 		SVC_DESTROY(xprt);
-#ifdef USE_LTTNG_NTIRPC
-		tracepoint(xprt, recv_exit, __func__, __LINE__,
-			   xprt, "REARM FAILED", -1);
-#endif /* USE_LTTNG_NTIRPC */
+
+		XPRT_UNIQUE_AUTO_TRACEPOINT(xprt, rearm_failed,
+			TRACE_ERR, "Rearm failed");
+
 		return SVC_STAT(xprt);
 	}
 
-#ifdef USE_LTTNG_NTIRPC
-	tracepoint(xprt, recv_exit, __func__, __LINE__,
-		   xprt, "CALLING svc_request", 0);
-#endif /* USE_LTTNG_NTIRPC */
+	XPRT_UNIQUE_AUTO_TRACEPOINT(xprt, calling_svc_request,
+		TRACE_DEBUG, "Calling svc_request");
+
 	return svc_request(xprt, xioq->xdrs);
 }
 
