@@ -51,6 +51,10 @@
 #include <misc/abstract_atomic.h>
 #include "rpc_com.h"
 
+#ifdef USE_RPC_RDMA
+#include "rpc_rdma.h"
+#endif
+
 #include <rpc/xdr_ioq.h>
 
 #define VREC_MAXBUFS 24
@@ -64,6 +68,32 @@ static uint64_t next_id;
 #endif				/* 0 */
 #define free_buffer(addr,size) mem_free((addr), size)
 
+#define NS_PER_SEC  ((uint64_t) 1000000000)
+/**
+ * @brief Get the abs difference between two timespecs in nsecs
+ *
+ * useful for cheap time calculation. Works with Dr. Who...
+ *
+ * @param[in] start timespec of before end
+ * @param[in] end   timespec after start time
+ *
+ * @return Elapsed time in nsecs
+ */
+static inline uint64_t
+timespec_diff(const struct timespec *start,
+              const struct timespec *end)
+{
+        if ((end->tv_sec > start->tv_sec)
+            || (end->tv_sec == start->tv_sec
+                && end->tv_nsec >= start->tv_nsec)) {
+                return (end->tv_sec - start->tv_sec) * NS_PER_SEC +
+                    (end->tv_nsec - start->tv_nsec);
+        } else {
+                return (start->tv_sec - end->tv_sec) * NS_PER_SEC +
+                    (start->tv_nsec - end->tv_nsec);
+        }
+}
+
 struct xdr_ioq_uv *
 xdr_ioq_uv_create(size_t size, u_int uio_flags)
 {
@@ -187,6 +217,369 @@ xdr_ioq_uv_recycle(struct poolq_head *ioqh, struct poolq_entry *have)
 	pthread_mutex_unlock(&ioqh->qmutex);
 }
 
+#ifdef USE_RPC_RDMA
+
+static struct rpc_io_bufs *
+get_parent_chunk(struct poolq_entry *have)
+{
+	struct xdr_ioq_uv *uv = IOQ_(have);
+	return uv->u.uio_u1;
+}
+
+static int
+chunk_ref_locked(struct poolq_entry *have)
+{
+	struct rpc_io_bufs *io_buf = get_parent_chunk(have);
+	RDMAXPRT *rdma_xprt = (RDMAXPRT *)io_buf->ctx;
+
+	if ((io_buf != rdma_xprt->first_io_buf) &&
+	    ((io_buf->type == IO_INBUF_DATA) ||
+	    (io_buf->type == IO_OUTBUF_DATA))) {
+		atomic_inc_uint64_t(&rdma_xprt->total_extra_buf_allocations);
+		clock_gettime(CLOCK_MONOTONIC_FAST,
+		    &rdma_xprt->last_extra_buf_allocation_time);
+	}
+
+	return atomic_inc_uint32_t(&io_buf->refs);
+}
+
+static void
+do_shrink(RDMAXPRT *rdma_xprt, struct rpc_io_bufs *io_buf)
+{
+	struct poolq_head *ioqh = NULL;
+
+	switch (io_buf->type) {
+		case IO_INBUF_HDR:
+			break;
+		case IO_INBUF_DATA:
+			ioqh = &rdma_xprt->inbufs_data.uvqh;
+			break;
+		case IO_OUTBUF_HDR:
+			break;
+		case IO_OUTBUF_DATA:
+			ioqh = &rdma_xprt->outbufs_data.uvqh;
+			break;
+		default:
+			assert(io_buf->type == IO_BUF_ALL);
+	}
+
+	assert(ioqh);
+
+	__warnx(TIRPC_DEBUG_FLAG_EVENT, "%s: Start shrinking xprt %p "
+	    "io_buf %p refs %d ioqh %p count %d",
+	    __func__, rdma_xprt, io_buf, io_buf->refs, ioqh,
+	    ioqh->qcount);
+
+	xdr_rdma_buf_pool_destroy_locked(ioqh, io_buf);
+}
+
+#define NSEC_IN_SEC (1000ULL*1000ULL*1000ULL)
+#define SHRINK_WAIT_TIME_NS (NSEC_IN_SEC * 60ULL)
+
+static struct rpc_io_bufs *
+get_lru_chunk(RDMAXPRT *rdma_xprt)
+{
+	struct timespec ts_end, ts_start;
+	uint64_t diff_ns = 0;
+	struct rpc_io_bufs *io_buf = NULL;
+
+	clock_gettime(CLOCK_MONOTONIC_FAST, &ts_end);
+	ts_start = rdma_xprt->last_extra_buf_allocation_time;
+
+	if (ts_end.tv_sec > ts_start.tv_sec)
+		diff_ns = timespec_diff(&ts_start, &ts_end);
+
+	if (diff_ns >= SHRINK_WAIT_TIME_NS) {
+		pthread_mutex_lock(&rdma_xprt->io_bufs.qmutex);
+		struct poolq_entry *have = TAILQ_FIRST(&rdma_xprt->io_bufs.qh);
+		while (have) {
+			io_buf = opr_containerof(have, struct rpc_io_bufs, q);
+
+			/* We are with ioqh_lock so no other thread can allocated
+			 * from thi ioqh and get io_buf ref */
+			if (io_buf->ready && (io_buf != rdma_xprt->first_io_buf) &&
+			    (io_buf->refs == 0) && ((io_buf->type == IO_INBUF_DATA) ||
+			    (io_buf->type == IO_OUTBUF_DATA)))
+				break;
+			struct poolq_entry *next = TAILQ_NEXT(have, q);
+			have = next;
+			io_buf = NULL;
+		}
+		if (io_buf) {
+			TAILQ_REMOVE(&rdma_xprt->io_bufs.qh, &io_buf->q, q);
+			rdma_xprt->io_bufs_count--;
+			rdma_xprt->io_bufs.qcount--;
+
+			__warnx(TIRPC_DEBUG_FLAG_EVENT, "%s: xprt %p shrink io_buf %p"
+			    "refs %d io_bufs count %d %d",
+			    __func__, rdma_xprt, io_buf, io_buf->refs,
+			    rdma_xprt->io_bufs_count, rdma_xprt->io_bufs.qcount);
+		}
+		pthread_mutex_unlock(&rdma_xprt->io_bufs.qmutex);
+	}
+
+	return io_buf;
+}
+
+static int
+chunk_unref_locked(struct poolq_entry *have)
+{
+	struct rpc_io_bufs *io_buf = get_parent_chunk(have);
+	uint32_t refs = atomic_dec_uint32_t(&io_buf->refs);
+	RDMAXPRT *rdma_xprt = (RDMAXPRT *)io_buf->ctx;
+
+	if ((io_buf != rdma_xprt->first_io_buf) &&
+	    ((io_buf->type == IO_INBUF_DATA) ||
+	    (io_buf->type == IO_OUTBUF_DATA)))
+		atomic_dec_uint64_t(&rdma_xprt->total_extra_buf_allocations);
+
+	io_buf = get_lru_chunk(rdma_xprt);
+	if (io_buf) {
+		do_shrink(rdma_xprt, io_buf);
+	}
+	return refs;
+}
+
+static bool
+parent_chunk(struct poolq_entry *have,
+    struct rpc_io_bufs *check_buf)
+{
+	struct rpc_io_bufs *io_buf = get_parent_chunk(have);
+	return (io_buf == check_buf);
+}
+
+struct poolq_entry *
+xdr_rdma_ioq_uv_fetch(struct xdr_ioq *xioq, struct poolq_head *ioqh,
+		 char *comment, u_int count, u_int ioq_flags)
+{
+	struct poolq_entry *have = NULL;
+	RDMAXPRT *rdma_xprt = NULL;
+
+	assert(xioq->rdma_ioq);
+
+	rdma_xprt = xioq->xdrs[0].x_lib[1];
+
+	__warnx(TIRPC_DEBUG_FLAG_XDR,
+		"%s() %u %s rdma_xprt %p",
+		__func__, count, comment, rdma_xprt);
+
+	pthread_mutex_lock(&ioqh->qmutex);
+
+	while (1) {
+		/* positive for buffer(s) */
+		have = TAILQ_FIRST(&ioqh->qh);
+		__warnx(TIRPC_DEBUG_FLAG_XDR,
+			"%s() %u %s xioq %p %d ioqh %p %d have %p",
+			__func__, count, comment, xioq, xioq->ioq_uv.uvqh.qcount,
+			ioqh, ioqh->qcount, have);
+		if (have) {
+			TAILQ_REMOVE(&ioqh->qh, have, q);
+			if (ioqh != &rdma_xprt->cbqh)
+				chunk_ref_locked(have);
+
+			/* added directly to the queue.
+			 * this lock is needed for context header queues,
+			 * but is not a burden on uncontested data queues.
+			 */
+			pthread_mutex_lock(&xioq->ioq_uv.uvqh.qmutex);
+			(xioq->ioq_uv.uvqh.qcount)++;
+			__warnx(TIRPC_DEBUG_FLAG_XDR, "ioq_track xdr_ioq_uv_fetch insert "
+				"have %p to q %p", have, xioq);
+			TAILQ_INSERT_TAIL(&xioq->ioq_uv.uvqh.qh, have, q);
+			pthread_mutex_unlock(&xioq->ioq_uv.uvqh.qmutex);
+			ioqh->qcount--;
+			count--;
+			if(count == 0)
+				break;
+		} else {
+			if (rdma_xprt) {
+				if (ioqh == &rdma_xprt->inbufs_data.uvqh) {
+					xdr_rdma_add_inbufs_data(rdma_xprt);
+				}
+
+				if (ioqh == &rdma_xprt->outbufs_data.uvqh) {
+					xdr_rdma_add_outbufs_data(rdma_xprt);
+				}
+
+				if (ioqh == &rdma_xprt->inbufs_hdr.uvqh) {
+					xdr_rdma_add_inbufs_hdr(rdma_xprt);
+				}
+
+				if (ioqh == &rdma_xprt->outbufs_hdr.uvqh) {
+					xdr_rdma_add_outbufs_hdr(rdma_xprt);
+				}
+
+				if (unlikely(ioqh == &rdma_xprt->cbqh)) {
+					__warnx(TIRPC_DEBUG_FLAG_EVENT, "cbc buffers exhausetd rdma_xprt %p "
+						"ioqh %p qcount %d", rdma_xprt, ioqh, ioqh->qcount);
+					rpc_rdma_allocate_cbc_locked(ioqh);
+				}
+
+			}
+		}
+	}
+
+	pthread_mutex_unlock(&ioqh->qmutex);
+	return have;
+}
+
+struct poolq_entry *
+xdr_rdma_ioq_uv_fetch_nothing(struct xdr_ioq *xioq, struct poolq_head *ioqh,
+			 char *comment, u_int count, u_int ioq_flags)
+{
+	return NULL;
+}
+
+static inline void
+xdr_rdma_ioq_uv_recycle(struct poolq_head *ioqh, struct poolq_entry *have)
+{
+	pthread_mutex_lock(&ioqh->qmutex);
+
+	__warnx(TIRPC_DEBUG_FLAG_XDR, "%s() ioq_track Reccycle ioqh %p %d have %p",
+		__func__, ioqh, ioqh->qcount, have);
+
+	TAILQ_INSERT_TAIL(&ioqh->qh, have, q);
+	ioqh->qcount++;
+
+	pthread_mutex_unlock(&ioqh->qmutex);
+}
+
+static inline void
+xdr_rdma_ioq_uv_recycle_io_buf(struct poolq_head *ioqh,
+    struct poolq_entry *have)
+{
+	pthread_mutex_lock(&ioqh->qmutex);
+
+	__warnx(TIRPC_DEBUG_FLAG_XDR, "%s() ioq_track Reccycle ioqh %p %d have %p",
+		__func__, ioqh, ioqh->qcount, have);
+
+	struct rpc_io_bufs *io_buf = get_parent_chunk(have);
+	RDMAXPRT *rdma_xprt = (RDMAXPRT *)io_buf->ctx;
+
+	if (io_buf == rdma_xprt->first_io_buf)
+		TAILQ_INSERT_HEAD(&ioqh->qh, have, q);
+	else
+		TAILQ_INSERT_TAIL(&ioqh->qh, have, q);
+	ioqh->qcount++;
+
+	if (ioqh != &rdma_xprt->cbqh)
+		chunk_unref_locked(have);
+
+	pthread_mutex_unlock(&ioqh->qmutex);
+}
+
+void
+xdr_rdma_ioq_uv_release(struct xdr_ioq_uv *uv)
+{
+	/* Reset vectors since we use it to UIO_REFER */
+	uv->u = uv->rdma_u;
+	uv->v = uv->rdma_v;
+	xdr_rdma_ioq_uv_recycle_io_buf(uv->u.uio_p1, &uv->uvq);
+}
+
+void
+xdr_rdma_ioq_release(struct poolq_head *ioqh, bool xioq_recycle,
+    struct xdr_ioq *xioq)
+{
+	struct poolq_entry *have = TAILQ_FIRST(&ioqh->qh);
+
+	/* release queued buffers */
+	while (have) {
+		assert(xioq->rdma_ioq);
+
+		__warnx(TIRPC_DEBUG_FLAG_XDR, "%s ioqh %p %d "
+		    "xioq %p have %p", __func__,
+		    ioqh, ioqh->qcount, xioq, have);
+
+		pthread_mutex_lock(&ioqh->qmutex);
+
+		TAILQ_REMOVE(&ioqh->qh, have, q);
+		(ioqh->qcount)--;
+
+		pthread_mutex_unlock(&ioqh->qmutex);
+
+		xdr_rdma_ioq_uv_release(IOQ_(have));
+
+		have = TAILQ_FIRST(&ioqh->qh);
+	}
+
+	assert(ioqh->qcount == 0);
+
+	/* Recycle cbc */
+	if (xioq && xioq->ioq_pool && xioq_recycle) {
+		xdr_rdma_ioq_uv_recycle(xioq->ioq_pool, &xioq->ioq_s);
+	}
+
+}
+
+static void
+xdr_rdma_ioq_uv_destroy(struct xdr_ioq_uv *uv)
+{
+	mem_free(uv, sizeof(*uv));
+}
+
+void
+xdr_rdma_buf_pool_destroy(struct poolq_head *ioqh,
+    struct rpc_io_bufs *io_buf)
+{
+	pthread_mutex_lock(&ioqh->qmutex);
+	xdr_rdma_buf_pool_destroy_locked(ioqh, io_buf);
+	pthread_mutex_unlock(&ioqh->qmutex);
+}
+
+void
+xdr_rdma_buf_pool_destroy_locked(struct poolq_head *ioqh,
+    struct rpc_io_bufs *io_buf)
+{
+	/* pool_head may not be initialized, so check for qcount */
+	if (ioqh->qcount && !TAILQ_EMPTY(&ioqh->qh)) {
+
+		assert(io_buf->refs == 0);
+
+		struct poolq_entry *have = TAILQ_FIRST(&ioqh->qh);
+
+		/* release queued buffers */
+		while (have) {
+			struct poolq_entry *next = TAILQ_NEXT(have, q);
+
+			__warnx(TIRPC_DEBUG_FLAG_XDR,
+			    "%s ioqh %p %d have %p",
+			    __func__, ioqh, ioqh->qcount, have);
+
+			if (io_buf && !parent_chunk(have, io_buf)) {
+				have = next;
+				continue;
+			}
+
+
+			TAILQ_REMOVE(&ioqh->qh, have, q);
+			(ioqh->qcount)--;
+
+			atomic_dec_uint64_t(&io_buf->buf_count);
+
+			xdr_rdma_ioq_uv_destroy(IOQ_(have));
+			have = next;
+		}
+		if (0 == atomic_fetch_uint64_t(&io_buf->buf_count)) {
+			RDMAXPRT *rdma_xprt = (RDMAXPRT *)io_buf->ctx;
+			assert(io_buf->mr);
+			assert(!xdr_rdma_dereg_mr(rdma_xprt, io_buf->mr,
+			    io_buf->buffer_aligned, io_buf->buffer_total));
+			io_buf->mr = NULL;
+
+			__warnx(TIRPC_DEBUG_FLAG_EVENT, "%s() Free xprt %p mr "
+			    "io_bufs %p size %u io_buf %p", __func__, rdma_xprt,
+			    io_buf->buffer_aligned, io_buf->buffer_total, io_buf);
+
+			assert(io_buf->buffer_aligned);
+			mem_free(io_buf->buffer_aligned, io_buf->buffer_total);
+			io_buf->buffer_aligned = NULL;
+		}
+	}
+}
+
+#endif
+
 void
 xdr_ioq_uv_release(struct xdr_ioq_uv *uv)
 {
@@ -489,6 +882,143 @@ xdr_ioq_getbytes(XDR *xdrs, char *addr, u_int len)
 	return (true);
 }
 
+#ifdef USE_RPC_RDMA
+
+/*
+ * Destroy rdma buffers
+ */
+static void
+xdr_ioq_destroy_internal_rdma(XDR *xdrs)
+{
+	__warnx(TIRPC_DEBUG_FLAG_XDR, "%s: no op for rdma",
+	    __func__);
+}
+
+/*
+ * Get start position for rdma data
+ */
+static u_int
+xdr_ioq_getstartdatapos_rdma(XDR *xdrs, u_int start, u_int datalen)
+{
+	u_int offset = 0;
+	struct xdr_ioq *xioq = XIOQ(xdrs);
+	struct xdr_ioq_uv *uv = IOQV(xioq->xdrs[0].x_base);
+
+	assert(xioq->rdma_ioq);
+	assert(ioquv_size(uv) == RDMA_HDR_CHUNK_SZ);
+
+	/* Check if data is not inline.
+	 * If data is inline, it should be part of nfs_buffer itself.
+	 * If data is not inline, then rdma_read buffers will be placed after
+	 * nfs_buffer, so calculate offset for end of nfs_buffer */
+	if (datalen > ((uintptr_t)xdrs->x_v.vio_tail - (uintptr_t)xdrs->x_data)) {
+		offset = (uintptr_t)xdrs->x_v.vio_tail - (uintptr_t)xdrs->x_data;
+	}
+
+	return start + offset;
+
+}
+
+/*
+ * Get end position for rdma data
+ */
+static u_int
+xdr_ioq_getenddatapos_rdma(XDR *xdrs, u_int start, u_int datalen)
+{
+	u_int offset = 0;
+	struct xdr_ioq *xioq = XIOQ(xdrs);
+	struct xdr_ioq_uv *uv = IOQV(xioq->xdrs[0].x_base);
+
+	assert(xioq->rdma_ioq);
+	assert(ioquv_size(uv) == RDMA_HDR_CHUNK_SZ);
+
+	/* Check if data is not inline.
+	 * If data is inline, it should be part of nfs_buffer itself.
+	 * If data is not inline, then rdma_read buffers will be placed after
+	 * nfs_buffer, so we need to set start within nfs_buffer itself to read
+	 * next nfs header in compound op. */
+	if (datalen > ((uintptr_t)xdrs->x_v.vio_tail - (uintptr_t)xdrs->x_data)) {
+		offset = (uintptr_t)xdrs->x_v.vio_tail - (uintptr_t)xdrs->x_data;
+	}
+
+	return start - offset;
+}
+
+static bool
+xdr_ioq_getbytes_rdma(XDR *xdrs, char *addr, u_int len)
+{
+	struct xdr_ioq_uv *uv;
+	ssize_t delta;
+	int pcount = 0;
+
+	struct xdr_ioq *xioq = XIOQ(xdrs);
+	XDR orig_xdr;
+	int restore_xdr = 0;
+
+	__warnx(TIRPC_DEBUG_FLAG_XDR, "enter %s: xdata %p xioq %p rdma %d len %u "
+	    "pcount %d qcount %d", __func__, xdrs->x_data, xioq, xioq->rdma_ioq,
+	    len, XIOQ(xdrs)->ioq_uv.pcount, XIOQ(xdrs)->ioq_uv.uvqh.qcount);
+
+	/* Check if we are getting rdma_write bytes, we could have some
+	 * header part to get next compound, so restore hdr xdr at end */
+	if (xioq->rdma_ioq &&
+	    (len > ((uintptr_t)xdrs->x_v.vio_tail - (uintptr_t)xdrs->x_data))) {
+
+		__warnx(TIRPC_DEBUG_FLAG_XDR, "%s: rdma_write len %u hdr delta %u",
+		    __func__, len, ((uintptr_t)xdrs->x_v.vio_tail -
+		    (uintptr_t)xdrs->x_data));
+
+		memcpy(&orig_xdr, xdrs, sizeof(XDR));
+		uv = xdr_ioq_uv_advance(XIOQ(xdrs));
+		if (!uv) {
+			__warnx(TIRPC_DEBUG_FLAG_XDR, "%s NULL uv", __func__);
+			return (false);
+		}
+		xdr_ioq_uv_update(XIOQ(xdrs), uv);
+		restore_xdr = 1;
+	}
+
+	while (len > 0
+		&& XIOQ(xdrs)->ioq_uv.pcount < XIOQ(xdrs)->ioq_uv.uvqh.qcount) {
+		delta = (uintptr_t)xdrs->x_v.vio_tail
+			- (uintptr_t)xdrs->x_data;
+
+		if (unlikely(delta > len)) {
+			delta = len;
+		} else if (unlikely(!delta)) {
+			/* advance fill pointer */
+			uv = xdr_ioq_uv_advance(XIOQ(xdrs));
+			if (!uv) {
+				return (false);
+			}
+			xdr_ioq_uv_update(XIOQ(xdrs), uv);
+			continue;
+		}
+		memcpy(addr, xdrs->x_data, delta);
+		xdrs->x_data += delta;
+		addr += delta;
+		len -= delta;
+	}
+
+	pcount = XIOQ(xdrs)->ioq_uv.pcount;
+
+	if (restore_xdr) {
+		__warnx(TIRPC_DEBUG_FLAG_XDR, "%s: rdma_write restore xdr hdr",
+		    __func__);
+		memcpy(xdrs, &orig_xdr, sizeof(XDR));
+		XIOQ(xdrs)->ioq_uv.pcount = pcount;
+	}
+
+	__warnx(TIRPC_DEBUG_FLAG_XDR, "exit %s: xdata %p xioq %p rdma %d len %u "
+	    "pcount %d qcount %d", __func__, xdrs->x_data, xioq, xioq->rdma_ioq,
+	    len, XIOQ(xdrs)->ioq_uv.pcount, XIOQ(xdrs)->ioq_uv.uvqh.qcount);
+
+	assert(len == 0);
+	return (true);
+}
+
+#endif
+
 static bool
 xdr_ioq_putbytes(XDR *xdrs, const char *addr, u_int len)
 {
@@ -681,6 +1211,24 @@ xdr_ioq_getpos(XDR *xdrs)
 		   - (uintptr_t)xdrs->x_v.vio_head));
 }
 
+/*
+ * Get position for start of data
+ */
+static u_int
+xdr_ioq_getstartdatapos(XDR *xdrs, u_int start, u_int datalen)
+{
+	return start;
+}
+
+/*
+ * Get position for end of data
+ */
+static u_int
+xdr_ioq_getenddatapos(XDR *xdrs, u_int start, u_int datalen)
+{
+	return start + datalen;
+}
+
 /*
  * Set read/insert or fill position.
  *
@@ -769,6 +1317,10 @@ xdr_ioq_destroy(struct xdr_ioq *xioq, size_t qsize)
 		"%s() xioq %p",
 		__func__, xioq);
 
+#ifdef USE_RPC_RDMA
+	assert(!xioq->rdma_ioq);
+#endif
+
 	xdr_ioq_release(&xioq->ioq_uv.uvqh);
 
 	if (xioq->ioq_pool) {
@@ -1261,6 +1813,8 @@ const struct xdr_ops xdr_ioq_ops = {
 	xdr_ioq_getbytes,
 	xdr_ioq_putbytes,
 	xdr_ioq_getpos,
+	xdr_ioq_getstartdatapos,
+	xdr_ioq_getenddatapos,
 	xdr_ioq_setpos,
 	xdr_ioq_destroy_internal,
 	xdr_ioq_control,
@@ -1271,3 +1825,24 @@ const struct xdr_ops xdr_ioq_ops = {
 	xdr_ioq_fillbufs,	/* x_fillbufs */
 	xdr_ioq_allochdrs,	/* x_allochdrs */
 };
+
+#ifdef USE_RPC_RDMA
+const struct xdr_ops xdr_ioq_ops_rdma = {
+	xdr_ioq_getunit,
+	xdr_ioq_putunit,
+	xdr_ioq_getbytes_rdma,
+	xdr_ioq_putbytes,
+	xdr_ioq_getpos,
+	xdr_ioq_getstartdatapos_rdma,
+	xdr_ioq_getenddatapos_rdma,
+	xdr_ioq_setpos,
+	xdr_ioq_destroy_internal_rdma,
+	xdr_ioq_control,
+	xdr_ioq_getbufs,
+	xdr_ioq_putbufs,
+	xdr_ioq_newbuf,		/* x_newbuf */
+	xdr_ioq_iovcount,	/* x_iovcount */
+	xdr_ioq_fillbufs,	/* x_fillbufs */
+	xdr_ioq_allochdrs,	/* x_allochdrs */
+};
+#endif
